import requests
import ujson as json


def _keys(nodes):
    """Returns a list of keys from the given nodes.

    eg.
    ---------------------------------------------------------------------------
        For a node like the below:
    ---------------------------------------------------------------------------
            [
                {
                    "createdIndex": 137,
                    "key": "/dev/xx.ABC001/authorizedKeys/yy.DEF001",
                    "modifiedIndex": 137,
                    "value": ""
                }, {
                    "createdIndex": 192,
                    "key": "/dev/xx.ABC001/authorizedKeys/zz.GHI001",
                    "modifiedIndex": 192,
                    "value": ""
                }
            ]
    ---------------------------------------------------------------------------
        The return would be:
    ---------------------------------------------------------------------------
            ["yy.DEF001", "zz.GHI001"]

    """
    keys = {}
    for node in nodes:
        if 'key' not in node:
            continue

        # The key is the value after the last `/`.
        # eg:
        #   for this value `/prod/ci.MPP001/privateKey`
        #   the key would be `privateKey`
        i = node['key'].rfind('/')+1
        key = node['key'][i:]
        # Append the key to the list.
        keys[key] = node['value']
    return keys

def _data(etcd_url):
    """Invokes the given URL, parses the JSON response, and returns
    either a node list, or the value of a node.

    ---------------------------------------------------------------------------
    # For the below etcd response:
    ---------------------------------------------------------------------------
        {
            "action": "get",
            "node": {
                "createdIndex": 132,
                "dir": true,
                "key": "/prod/ci.USR001/authorizedKeys",
                "modifiedIndex": 132,
                "nodes": [
                    {
                      "createdIndex": 137,
                      "key": "/prod/ci.USR001/authorizedKeys/ci.PCP001",
                      "modifiedIndex": 137,
                      "value": ""
                    },
                    {
                      "createdIndex": 192,
                      "key": "/prod/ci.USR001/authorizedKeys/ci.SIT001",
                      "modifiedIndex": 192,
                      "value": ""
                    }
                ]
            }
        }
    ---------------------------------------------------------------------------
    # The nodes list would be returned:
    ---------------------------------------------------------------------------
        "nodes": [
            {
              "createdIndex": 137,
              "key": "/prod/ci.USR001/authorizedKeys/ci.PCP001",
              "modifiedIndex": 137,
              "value": ""
            },
            {
              "createdIndex": 192,
              "key": "/prod/ci.USR001/authorizedKeys/ci.SIT001",
              "modifiedIndex": 192,
              "value": ""
            }
        ]
    ---------------------------------------------------------------------------
    # For the below etcd response:
    ---------------------------------------------------------------------------
        {
            "action": "get",
            "node": {
                "createdIndex": 133,
                "key": "/prod/ci.USR001/privateKey",
                "modifiedIndex": 133,
                "value": "b2c30cfd-4f34-4c2d-9af0-c69fe1eb338a"
            }
        }
    ---------------------------------------------------------------------------
    # The value would be returned:
    ---------------------------------------------------------------------------
        "b2c30cfd-4f34-4c2d-9af0-c69fe1eb338a"

    """
    r = requests.get(etcd_url)
    if r.status_code != 200:
        err = 'Invalid status {status} from etcd.'.format(status=r.status_code)
        raise Exception(err)

    j = r.json()
    if 'node' not in j:
        raise Exception('"node" missing from etcd response.')

    node = j['node']
    # Either have a collection of nodes or a value.
    if 'nodes' in node:
        return _keys(node['nodes'])
    elif 'value' not in node:
        raise Exception('"value" missing from etcd response node.')
    return node['value']

def authorized_key(etcd_url, key):
    """Gets the private key matching the given key from etcd.

    This is used to grab a private key that has been hard-set within
    an entries `authorizedKeys` collection.
    """
    return _data('{0}/authorizedKeys/{1}'.format(etcd_url, key))

def authorized_keys(etcd_url):
    """Returns a list of authorized keys for the entry.
    """
    return _data(etcd_url+'/authorizedKeys')

def private_key(etcd_url):
    """Returns the private key for the entry.
    """
    return _data(etcd_url+'/privateKey')

def config(etcd_url):
    """Returns the configuration for the entry.
    """
    data = _data(etcd_url+'/configuration')
    if data:
        return json.loads(data)
    return {}



import json, sys, time
from os import environ

from flask import Flask

import tinflask.etcd
import tinflask.keystore
import tinflask.handlers


class Application(object):
    """tinflask.web.Application instance.

    The built-in `/status` endpoint now requires authorization.
    Keys endpoints are also added to hot-reload authorized/private keys, and
    simple listing of loaded public keys.

    A new instance will set the global key function within tinflask.handlers for
    all signed/authorized endpoints. This key function retrieves private keys
    for the authorization process.
    """

    def __init__(self, handlers=None, **settings):
        """Creates a new tinflask/app with the given handlers and settings.

        - If an `ENDPOINT` environment variable is set all handlers will be prefixed with it.
        - The `status` endpoint is re-added as a signed endpoint.
        - Key endpoints `keys`, `keys/private`, and `keys/authorized` are also added as signed endpoints.
        """

        # Composite app.
        self.app = Flask(__name__)

        # Set the port, defaulting to 8000
        try:
            port_str = environ.get("PORT", "8080")
            self.port = int(port_str)
        except ValueError:
            self.port = 8080

        # Set the host and endpoint
        self.debug = True if environ.get("DEBUG") == "true" else False
        self.address = environ.get("HOST", "0.0.0.0")
        self.jobs = []

        # Add standard endpoints.
        self.app.route("/ping")(tinflask.handlers.ping)
        self.app.route("/time")(tinflask.handlers.time)
        status_handler = tinflask.handlers.status(self.jobs)
        self.app.route("/status")(status_handler)

    def add_status_job(self, name, timeout, job_func):
        """Adds the given status job to the collection of jobs.
        """
        self.jobs.append((name, timeout, job_func,))

    def run(self):
        """Starts the Flask service using the address, port, and debug values
        found from either environment variables or config variables.
        """
        self.app.run(host=self.address, port=self.port, debug=self.debug)



import concurrent.futures
import functools
from time import time as now
import sys
import time
from timeit import timeit

import ujson as json
from flask import request, Response

import hancock


def signed(key_func, log_func=sys.stdout.write, expire_seconds=10):
    """Validates the signature of the request before invoking the handler.
    Logging is passed by default to stdout.
    If the signature fails a 401 response is sent.
    If the signature has expired a 406 response is sent.

    key_func       - A function that takes a public key as a single argument and
                     returns the matching private key.
    log_func       - A function taking a single argument to be logged.
    expire_seconds - Integer of how many seconds before authorization is no
                     longer valid.
    """
    def signed_decorator(fn):
        @functools.wraps(fn)
        def wrapper(*args, **kwargs):
            # Get the public key
            key = request.args.get("apikey")
            if not key:
                return '', 401
            # Get the private key for the public key.
            p_key = key_func(key)
            if not p_key:
                log_func("No private key for: {key}".format(key=key))
                return '', 401

            # Validate the request.
            method = request.method
            query = dict(request.args)
            status_code, message = hancock.validate(p_key, query, expire_seconds, method)
            if status_code != 200:
                log_func(message)
                return '', status_code

            # Invoke the handler
            return fn(*args, **kwargs)

        return wrapper
    return signed_decorator


def status(jobs):
    """Handler that calls each status job in a worker pool, attempting to timeout.
    The resulting durations/errors are written to the response
    as JSON.

    eg.

    `{
        "endpoints": [
            { "endpoint": "Jenny's Database", "duration": 1.002556324005127 },
            { "endpoint": "Hotmail", "duration": -1, "error": "route to host down" },
        ]
     }`
    """
    def status_handler():
        endpoints = []
        stats = { "endpoints": None }

        executor = concurrent.futures.ThreadPoolExecutor(max_workers=5)
        # This is basically calling the below within the executor:
        #
        #     >>> timeit(job[2], number=1)
        #
        # job is a tuple of (name, timeout, func) so the above is really:
        #     >>> timeit(func, number=1)
        #
        for job, future in [(job, executor.submit(timeit, job[2], number=1)) for job in jobs]:
            name, timeout, _ = job
            endpoint = {"endpoint": name}
            try:
                data = future.result(timeout=timeout)
                endpoint["duration"] = data
            except concurrent.futures.TimeoutError:
                endpoint["error"] = "timeout exceeded"
            except Exception as ex:
                endpoint["error"] = str(ex)
            endpoints.append(endpoint)

        if len(endpoints) > 0:
            stats["endpoints"] = endpoints

        json_data = json.dumps(stats)
        resp = Response(response=json_data,
                status=200,
                mimetype="application/json")
        executor.shutdown(wait=False)
        return resp
        # TODO: Look into potentially cleaning up jobs that have timed-out.
        #
        #       This could be done by changing jobs from a func to an object
        #       that implements `def interrupt(self):` which would be used
        #       to interrupt/stop/close/cleanup any resources.
    return status_handler


def ping():
    """Handler that simply returns `pong` from a GET.
    """
    return "pong"


def time():
    """Handler that returns the systems current Epoch time.
    """
    utcnow = now()
    return str(int(utcnow))


def keys(key_store):
    def keys_decorator(fn):
        @wraps(fn)
        def handler(*args, **kwargs):
            key_store.load_keys()
        return handler
    return keys_decorator

def private_keys(key_store):
    def private_keys_decorator(fn):
        @wraps(fn)
        def handler(*args, **kwargs):
            key_store.load_private_key()
        return handler
    return private_keys_decorator

def authorized_keys(key_store):
    def authorized_keys_decorator(fn):
        @wraps(fn)
        def handler(*args, **kwargs):
            if request.method == "GET":
                keys = key_store.keys()
                return json.dumps(keys)
            elif request.method == "PUT":
                key_store.load_keys()
        return handler
    return authorized_keys_decorator



from tinflask import etcd


class KeyStore(object):
    """Store/cache for public/private key pairs for endpoint signing.
    """

    def __init__(self, key, config):
        """Creates a new key store for the given key and configuration.
        """
        self.key = key
        self.private_key = None
        self._keys = {}
        self._config = config

    
    def has_key(self, key):
        """Returns wether the given key is present in the store.
        """
        return key in self._keys

    def keys(self):
        """Returns a list of all keys within the store.
        """
        return list(self._keys)

    def get(self, key):
        """Gets the private key for the given key.
        """
        return self._keys.get(key)

    def load_keys(self):
        """Not Implemented.
        """
        pass

    def load_private_key(self):
        """Not Implemented.
        """
        pass


class File(KeyStore):
    """Configuration file backed KeyStore.

    load functions are side effect free.
    """

    def __init__(self, key, config):
        """Reads private and public keys from configuration file based on environment.
        """
        super(File, self).__init__(key, config)

        self._keys = self._config.get("authorizedKeys", {})
        self.private_key = self._config.get("privateKey")


class ETCD(KeyStore):
    """Etcd backed KeyStore.
    """

    def __init__(self, key, config, etcd_url):
        """Creates a new etcd backed KeyStore.
        """
        super(ETCD, self).__init__(key, config)

        # Load both the private and public keys.
        self._etcd_url = etcd_url
        self.load_private_key()
        self.load_keys()

    def get_etcd(self, key):
        """Returns the the private key from etcd directly.
        """
        url = "%s/%s" % (self._etcd_url, key)
        private_key = etcd.private_key(url)
        return private_key

    def get(self, key):
        """Lazy loads the matching private_key for the given public key.
        """
        private_key = self._keys.get(key, -1)
        if private_key == -1:   # Public key doesn't exist.
            return None
        elif private_key != '': # Private key is already set, return it.
            return private_key

        # Lazy load the private key from etcd.
        private_key = self.get_etcd(key)
        self._keys[key] = private_key
        return private_key

    def load_keys(self):
        """Loads authorized public keys from etcd.
        """
        url = "%s/%s" % (self._etcd_url, self.key)
        try:
            self._keys = etcd.authorized_keys(url)
        except Exception as ex:
            #TODO log this
            print(ex)

    def load_private_key(self):
        """Loads private key from etcd.
        """
        url = "%s/%s" % (self._etcd_url, self.key)
        try:
            self.private_key = etcd.private_key(url)
        except Exception as ex:
            #TODO log this
            print(ex)



import os
from setuptools import setup

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

setup(
    name = "tinflask",
    version = "0.0.1",
    author = "Justin Wilson",
    author_email = "justinwilson1@gmail.com",
    description = ("Simple wrapper around flask that uses "
        "environment variables for host, port, endpoint prefix. "
        "Also uses the py-hancock library for the ability to sign "
        "endpoints. Endpoints for `time`, `ping`, and `status` are "
        "automatically added as well."),
    license = "Simplified BSD",
    url = "http://pypi.python.org/pypi/tinflask",
    packages=['tinflask',],
    long_description=read('README.rst'),
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Topic :: Utilities",
        "License :: OSI Approved :: BSD License",
        "Natural Language :: English",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3.2",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
    ],
    install_requires=[
        "flask >= 0.10.1",
    	"requests >= 2.3.0",
    ],
)
